# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import torch.linalg
import torch
import torch._dynamo
from typing import List, Optional, Union, Mapping, Any
from functorch.compile import make_boxed_func
import ttnn
import pickle
from pathlib import Path
import os
from torch_ttnn.handle_input_aliasing import insert_clones_for_input_aliasing
import tools.export_code as export_code
import torch_ttnn.metrics as metrics
from torch_ttnn import mem_utils
from torch_ttnn.utils import GraphCleanup
import copy
from torch_ttnn.utils import get_add_custom_object_in_graph
import logging

torch._dynamo.config.suppress_errors = False
torch._dynamo.config.verbose = True


# The option for torch_ttnn.backend
class TorchTtnnOption:
    def __init__(
        self,
        device: ttnn.Device,
        gen_graphviz=False,
        run_mem_analysis=False,
        run_eviction_opt=False,
        verbose=True,
        metrics_path="",
        tracer_option=None,
        bypass_compile=False,
        use_less_ttnn_op_types=True,
        export_code=None,
        total_num_iterations=1,
        data_parallel=False,
        load_params_once=True,
        native_integration=False,
    ):
        self.device = device
        self.gen_graphviz = gen_graphviz
        self._out_fx_graphs = list()
        self.memory_manager = None
        self.run_mem_analysis = run_mem_analysis
        self.run_eviction_opt = run_eviction_opt
        self.verbose = verbose
        self.tracer_option = tracer_option
        self.total_num_iterations = total_num_iterations
        self.data_parallel = data_parallel

        self.metrics_path = metrics_path
        self.bypass_compile = bypass_compile
        self.use_less_ttnn_op_types = use_less_ttnn_op_types
        self.original_schema_list = list()
        self.compiled_schema_list = list()

        # Used for generate standalone python script
        self.export_code = export_code
        self._aten_fx_graphs = list()
        self._all_inputs = list()
        self._ttnn_fx_graphs = list()

        # Used for multi-device
        self._n_parameters = None
        self._n_buffers = None
        self._n_arguments = None

        # Used for pre-loading model params
        self._is_end_to_end = False

        self.load_params_once = load_params_once

        self.native_integration = native_integration

    def reset_containers(self):
        self._out_fx_graphs = list()
        self.original_schema_list = list()
        self._ttnn_fx_graphs = list()


def register_ttnn_objects(option: TorchTtnnOption):
    """
    torch.fx builds a source object as a string, calls builtin compile(), and finally
    calls builtin exec() with a dictionary of globals that contains the strings (keys)
    that will be replaced by the ttnn objects (values) during evaluation.
    """
    get_add_custom_object_in_graph("ttnn_Specified_Device", option.device)

    get_add_custom_object_in_graph("ttnn_ROW_MAJOR_LAYOUT", ttnn.ROW_MAJOR_LAYOUT)
    get_add_custom_object_in_graph("ttnn_TILE_LAYOUT", ttnn.TILE_LAYOUT)

    get_add_custom_object_in_graph("ttnn_uint32", ttnn.uint32)
    get_add_custom_object_in_graph("ttnn_int32", ttnn.int32)
    get_add_custom_object_in_graph("ttnn_bfloat16", ttnn.bfloat16)

    get_add_custom_object_in_graph("ttnn_DRAM_MEMORY_CONFIG", ttnn.DRAM_MEMORY_CONFIG)
    get_add_custom_object_in_graph("ttnn_L1_MEMORY_CONFIG", ttnn.L1_MEMORY_CONFIG)


# The backend for torch.compile that converts a graph to use ttnn.
# The "option" parameter is a TorchTtnnOption object
# See document for detail.
def aten_backend(
    gm: torch.fx.GraphModule,
    example_inputs: List[torch.Tensor],
    options: TorchTtnnOption,
) -> torch.fx.GraphModule:
    """
    The backend for torch.compile that converts a graph to use ttnn.
    The graph is wrapped in torch._dynamo.backends.common.aot_autograd, which
    trace into low level ATen ops not only high level torch ops.
    """

    option: TorchTtnnOption = options

    # Clone ops used for input aliasing workaround are no longer needed at this point
    from .handle_input_aliasing import remove_clones_for_input_aliasing

    gm = remove_clones_for_input_aliasing(gm)

    # Save aten graph if requested
    if options.export_code:
        # Will this hamper memory usage?
        graph_copy = copy.deepcopy(gm.graph)
        graph_copy.owning_module = gm
        option._aten_fx_graphs[-1].append(graph_copy)

    # Save the number of aten ops before compilation
    if option.metrics_path:
        # do constant folding for consistency of input varation
        from torch.fx.passes.infra.pass_manager import PassManager
        from torch_ttnn.passes.constant_folding_pass import ConstantFoldingPass

        pm_fold = PassManager(passes=[ConstantFoldingPass()])
        gm_fold, modified = pm_fold(gm)
        option.original_schema_list.extend(metrics.collect_input_variations_from_list_nodes(gm_fold.graph.nodes))

    # Do not continue with compilation if bypass
    if option.bypass_compile:
        option._out_fx_graphs.append(gm.graph)
        return make_boxed_func(gm)

    # Register ttnn objects as graph globals
    register_ttnn_objects(option)

    # Run analysis passes to help with ttnn ops
    from torch.fx.passes.infra.pass_manager import PassManager
    from torch_ttnn.passes.analysis.graph_module_analysis_pass import GraphModuleAnalysisPass
    from torch_ttnn.passes.analysis.input_analysis_pass import InputAnalysisPass
    from torch_ttnn.passes.analysis.multi_device_shard_analysis_pass import MultiDeviceShardAnalysisPass

    # Rewrite with ttnn ops, will insert redundant data movement
    from torch.fx.passes.dialect.common.cse_pass import CSEPass
    from torch_ttnn.passes.multi_device_pass import MultiDevicePass
    from torch_ttnn.passes.constant_folding_pass import ConstantFoldingPass
    from torch_ttnn.passes.lowering.to_tt_pass import ToTtPass
    from torch_ttnn.passes.fusion_pass import FusionPass
    from torch_ttnn.passes.lowering.add_data_move_pass import AddDataMovePass
    from torch_ttnn.passes.lowering.load_once_pass import LoadOncePass
    from torch_ttnn.passes.lowering.eliminate_coreops_pass import EliminateCoreopsPass
    from torch_ttnn.passes.graphviz_pass import GraphvizPass
    from torch_ttnn.passes.lowering.permute_reshape_tuple import PermuteReshapeTuple
    from torch_ttnn.passes.memory_pass import MemoryPass
    from torch_ttnn.passes.deallocation_pass import DeallocationPass

    passes = [
        GraphModuleAnalysisPass(),
        InputAnalysisPass(option._n_parameters, option._n_buffers, option._n_arguments),
        MultiDeviceShardAnalysisPass(option.device),
        ConstantFoldingPass(),
        MultiDevicePass(option.device, example_inputs),
        ToTtPass(option.device, option.use_less_ttnn_op_types),
        FusionPass(),
        AddDataMovePass(option.device),
        LoadOncePass(option._is_end_to_end, option.load_params_once),
        EliminateCoreopsPass(),
        CSEPass(),
        PermuteReshapeTuple(),
        DeallocationPass(),
    ]

    mem_pass = MemoryPass(option.verbose)
    if option.run_mem_analysis:
        passes.append(mem_pass)

    # Add graphviz pass interleavly if needed
    if option.gen_graphviz:
        passes_with_graphviz = [GraphvizPass(f"metrics/{option.metrics_path}/00.origin")]
        for idx in range(len(passes)):
            passes_with_graphviz.append(passes[idx])
            passes_with_graphviz.append(
                GraphvizPass(f"metrics/{option.metrics_path}/{idx + 1:02d}.{passes[idx].__class__.__name__}")
            )
        passes = passes_with_graphviz

    pm = PassManager(passes=passes)
    try:
        gm, modified = pm(gm)
    except Exception as e:
        logging.error(e.__context__)
        raise e

    GraphCleanup(gm)

    # Get the memory manager object for memory analysis
    if option.run_mem_analysis:
        option.memory_manager = mem_pass.mm

    # Run eviction opt pass if enabled
    if option.run_eviction_opt == True:
        assert option.run_mem_analysis == True, "Eviction pass depends on memory analysis pass!"
        from torch_ttnn.passes.eviction_pass import EvictionPass

        nth_eviction = 1
        max_evictions_limit = option.memory_manager.max_evictions_required()

        # See if SRAM overflows and eviction is required
        while mem_utils.check_sram_overflow(option.memory_manager) is True:
            if nth_eviction > max_evictions_limit:
                assert False, "Max evictions done, still model doesn't fit in memory!"

            guilty_op, tensors_to_evict = mem_utils.which_tensors_to_evict(option.memory_manager)
            # This indicates splitting is required
            if tensors_to_evict == -1:
                break
            # Run a eviction pass to remove tensors from device
            # This pass only evicts tensors for a single ttnn op which overflows the SRAM
            # Multiple overflows require multiple run of this pass

            mem_pass = MemoryPass(option.verbose)
            passes = [
                EvictionPass(option.memory_manager, guilty_op, tensors_to_evict),
                GraphvizPass(f"eviction-pass-{nth_eviction}"),
                mem_pass,
            ]
            pm = PassManager(passes=passes)
            gm, modified = pm(gm)

            GraphCleanup(gm)

            # Get the memory manager object for memory analysis
            option.memory_manager = mem_pass.mm
            nth_eviction += 1

    if option.metrics_path:
        option.compiled_schema_list.extend(metrics.collect_input_variations_from_list_nodes(gm.graph.nodes))

    option._out_fx_graphs.append(gm.graph)
    if options.export_code:
        option._ttnn_fx_graphs[-1].append(gm.graph)

    for node in gm.graph.nodes:
        if node.op == "placeholder":
            print(f"{node.name}: {node.meta}")

    for number, line in enumerate(gm.code.splitlines()):
        print(f"{number + 1}: {line}")

    return make_boxed_func(gm)


from torch._dynamo.backends.common import aot_autograd
from functools import partial


# The backend for torch.compile that converts a graph to use ttnn.
# The "option" parameter is a TorchTtnnOption object
# See document for detail.
# This function is a wrapper of aten_backend, and is used by torch.compile.
# This function is also registered as a backend for torch.compile.
def ttnn_backend(
    gm: torch.fx.GraphModule,
    example_inputs: List[torch.Tensor],
    options: TorchTtnnOption = None,
) -> torch.fx.GraphModule:
    if options.export_code:
        import tools.export_code as export_code

        # Some models have multiple forward functions with separate inputs for each.
        # Within these forward functions, there can be graph breakages which are
        # also represented by separate forward functions, but these do not have their
        # own separate inputs. Therefore, we organize the list of aten/ttnn graphs
        # with sublists where the top level list corresponds to the respective list of inputs.
        options._aten_fx_graphs.append(list())
        options._ttnn_fx_graphs.append(list())
        options._all_inputs.append(export_code.generate_flat_args(gm, example_inputs))

    # Analysis of params, buffers, and args
    options._n_parameters = len(list(gm.parameters()))
    options._n_buffers = len(list(gm.buffers()))
    options._n_arguments = len(example_inputs)

    # Currently, we only support preprocessing weights for end-to-end converted models
    options._is_end_to_end = gm.compile_subgraph_reason.graph_break == False

    tracer_option = options.tracer_option
    if tracer_option is not None:
        from ..tracer import Tracer

        out_prefix = f"fw_{tracer_option['model_name']}"
        out_folder = tracer_option["out_folder"]
        trace_orig = tracer_option["trace_orig"] if "trace_orig" in tracer_option else True
        trace_modi = tracer_option["trace_modi"] if "trace_modi" in tracer_option else False
        fw_compiler = Tracer(
            partial(aten_backend, options=options),
            out_prefix,
            out_folder,
            trace_orig,
            trace_modi,
        )
        return aot_autograd(fw_compiler=fw_compiler)(gm, example_inputs)
    else:
        gm = insert_clones_for_input_aliasing(gm)
        return aot_autograd(fw_compiler=partial(aten_backend, options=options))(gm, example_inputs)



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import torch
import ttnn

from torch.fx.passes.infra.pass_base import PassBase, PassResult


class PermuteReshapeTuple(PassBase):
    def call(self, gm: torch.fx.GraphModule):
        modified = False
        for node in gm.graph.nodes:
            # Only process ttnn.permute and ttnn.reshape
            if node.op != "call_function":
                continue
            if node.target not in [ttnn.permute, ttnn.reshape]:
                continue
            if isinstance(node.args[1], list):
                args = list(node.args)
                args[1] = tuple(node.args[1])
                args = tuple(args)
                node.args = args
                modified = True

        return PassResult(gm, modified)



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass
from operator import getitem
import ttnn
import torch
import pickle

from torch_ttnn.utils import TtnnDevice

run_once_count = 0
run_once_ans = tuple()


@torch.fx.wrap
@dataclass(frozen=True)
class RunOnceIdx:
    idx: int


@torch.fx.wrap
def run_once(*args):
    global run_once_count
    global run_once_ans

    if run_once_count == 0:
        temp_results = []
        return_results = []
        to_deallocate = []
        temp_idx_to_return_idx = dict()

        def lookup_function(str_name):
            # assume function is of form `ttnn.from_torch`, look up in globals()
            parts = str_name.split(".")
            base = globals()[parts[0]]
            for part in parts[1:]:
                base = getattr(base, part)
            return base

        def lookup_prev_result_idx(maybe_pickled_run_once_idx):
            try:
                run_once_idx = pickle.loads(maybe_pickled_run_once_idx)
                return run_once_idx.idx
            except pickle.UnpicklingError:
                pass
            return None

        def rewrite_args(arg_tuple):
            args = list(arg_tuple)
            for i, arg in enumerate(args):
                if isinstance(arg, (tuple, list, torch.fx.immutable_collections.immutable_list)):
                    original_version = list(arg)
                    decoded_version = [
                        lookup_prev_result_idx(entry) if isinstance(entry, bytes) else None
                        for entry in original_version
                    ]
                    decoded_version = [temp_results[idx] if idx is not None else None for idx in decoded_version]
                    new_arg = [
                        decoded if decoded is not None else original
                        for decoded, original in zip(decoded_version, original_version)
                    ]
                    args[i] = new_arg
                elif isinstance(arg, bytes):
                    maybe_prev_result_idx = lookup_prev_result_idx(arg)
                    if maybe_prev_result_idx is not None:
                        prev_val = temp_results[maybe_prev_result_idx]
                        args[i] = prev_val
            return tuple(args)

        def rewrite_kwargs(kwargs_dict):
            for k, v in kwargs_dict.items():
                if isinstance(v, (tuple, list, torch.fx.immutable_collections.immutable_list)):
                    original_version = list(v)
                    decoded_version = [
                        lookup_prev_result_idx(entry) if isinstance(entry, bytes) else None
                        for entry in original_version
                    ]
                    decoded_version = [temp_results[idx] if idx is not None else None for idx in decoded_version]
                    new_value = [
                        decoded if decoded is not None else original
                        for decoded, original in zip(decoded_version, original_version)
                    ]
                    kwargs_dict[k] = new_value
                elif isinstance(v, bytes):
                    maybe_prev_result_idx = lookup_prev_result_idx(v)
                    if maybe_prev_result_idx is not None:
                        prev_val = temp_results[maybe_prev_result_idx]
                        kwargs_dict[k] = prev_val
            return kwargs_dict

        def convert_input(spec):
            should_return, func_name, args, kwargs = spec
            found_func = lookup_function(func_name)
            # convert any args that reference previous results
            new_args = rewrite_args(args)
            kwargs = rewrite_kwargs(kwargs)
            temp_results.append(found_func(*new_args, **kwargs))
            if should_return:
                return_results.append(temp_results[-1])
                temp_idx_to_return_idx[len(temp_results) - 1] = len(return_results) - 1
            elif found_func == conv:
                # special case conv to preprocess weights and optional bias
                (dummy_output, (new_weights, new_bias)) = temp_results[-1]
                ttnn.deallocate(dummy_output)

                if (weight_idx := lookup_prev_result_idx(args[1])) is not None:
                    return_idx = temp_idx_to_return_idx[weight_idx]
                    return_results[return_idx] = new_weights
                    to_deallocate.append(weight_idx)

                bias_arg = args[2]
                if bias_arg is not None:
                    if (bias_idx := lookup_prev_result_idx(bias_arg)) is not None:
                        return_idx = temp_idx_to_return_idx[bias_idx]
                        return_results[return_idx] = new_bias
                        to_deallocate.append(bias_idx)
            else:
                to_deallocate.append(len(temp_results) - 1)

        for function_call in args:
            convert_input(function_call)

        # deallocate temporaries unless they alias a return value
        # this can happen when a returned value is calculated from a temp, but the operation doesn't actually do anything (e.g. to_layout(temp, ttnn.TILE_LAYOUT) but temp is already TILE_LAYOUT)
        # we only need to worry about device tensors here
        returned_addresses = [
            r.buffer_address()
            for r in return_results
            if isinstance(r, ttnn.Tensor) and r.storage_type() == ttnn.StorageType.DEVICE
        ]
        # only deallocate ttnn tensors
        to_deallocate = [temp_results[idx] for idx in to_deallocate if isinstance(temp_results[idx], ttnn.Tensor)]
        # only deallocate device tensors that do not alias a returned tensor
        to_deallocate = filter(
            lambda tens: tens.storage_type == ttnn.StorageType.DEVICE
            and tens.buffer_address() not in returned_addresses,
            to_deallocate,
        )
        for temp_tensor in to_deallocate:
            ttnn.deallocate(temp_tensor)

        run_once_ans = tuple(return_results)
        run_once_count += 1

    return run_once_ans


@torch.fx.wrap
def clone(t):
    return ttnn.clone(t, memory_config=t.memory_config(), dtype=t.dtype)


@torch.fx.wrap
def repeat(t, sizes):
    return ttnn.repeat(t, ttnn.Shape(sizes))


# Helper function to pack multiple values into a tuple on the graph
@torch.fx.wrap
def pack_to_tuple(*args):
    return tuple(args)


@torch.fx.wrap
def move_to_host(device_tensor, layout):
    device_tensor = ttnn.to_layout(device_tensor, layout)
    return ttnn.from_device(device_tensor)


@torch.fx.wrap
def conv(
    input_tensor,
    weight_tensor,
    bias_tensor,
    batch_size,
    in_channels,
    out_channels,
    in_spatial_shape,
    kernel_spatial_shape,
    stride,
    padding,
    dilation,
    groups,
    device,
    transposed,
    output_padding=None,
    return_weights_and_bias=False,
):
    if len(in_spatial_shape) == 1:
        # TODO(tt-metal#16258): conv1d API doesn't support transposed yet
        assert not transposed, "conv1d doesn't support transposed yet"
        return ttnn.conv1d(
            input_tensor=input_tensor,
            weight_tensor=weight_tensor,
            bias_tensor=bias_tensor,
            batch_size=batch_size,
            in_channels=in_channels,
            out_channels=out_channels,
            input_length=in_spatial_shape[0],
            kernel_size=kernel_spatial_shape[0],
            stride=stride[0],
            padding=padding[0],
            dilation=dilation[0],
            groups=groups,
            device=device,
            return_weights_and_bias=return_weights_and_bias,
        )
    if len(in_spatial_shape) == 2:
        in_h, in_w = in_spatial_shape
        if transposed:
            return ttnn.conv_transpose2d(
                input_tensor=input_tensor,
                weight_tensor=weight_tensor,
                bias_tensor=bias_tensor,
                batch_size=batch_size,
                in_channels=in_channels,
                out_channels=out_channels,
                input_height=in_h,
                input_width=in_w,
                kernel_size=kernel_spatial_shape,
                stride=stride,
                padding=padding,
                output_padding=output_padding,
                dilation=dilation,
                groups=groups,
                device=device,
                return_weights_and_bias=return_weights_and_bias,
            )
        else:
            assert output_padding is None, "conv2d has no output padding"
            return ttnn.conv2d(
                input_tensor=input_tensor,
                weight_tensor=weight_tensor,
                bias_tensor=bias_tensor,
                batch_size=batch_size,
                in_channels=in_channels,
                out_channels=out_channels,
                input_height=in_h,
                input_width=in_w,
                kernel_size=kernel_spatial_shape,
                stride=stride,
                padding=padding,
                dilation=dilation,
                groups=groups,
                device=device,
                return_weights_and_bias=return_weights_and_bias,
            )
    assert False, "unsupported conv shape"


@torch.fx.wrap
def roll(tensor, input_shape, shifts, dims):
    rolled_tensor = tensor
    for shift, dim in zip(shifts, dims):
        # slice tensor into two parts and concat them in reverse order
        end = (input_shape[dim] - shift) % input_shape[dim]

        # part1 = tensor[..., :end]
        slice_start, slice_end = [0] * len(input_shape), list(input_shape)
        slice_end[dim] = end
        sub_tensor1 = ttnn.slice(rolled_tensor, slice_start, slice_end)

        # part2 = tensor[..., end:]
        slice_start, slice_end = [0] * len(input_shape), list(input_shape)
        slice_start[dim] = end
        sub_tensor2 = ttnn.slice(rolled_tensor, slice_start, slice_end)

        # concat([part2, part1], dim)
        rolled_tensor = ttnn.concat([sub_tensor2, sub_tensor1], dim)

    return rolled_tensor


@torch.fx.wrap
def stack(tensors, dim, output_shape):
    # Handle negative dims by wrapping around
    dim = (dim + len(output_shape)) % len(output_shape)

    # Create shape for unsqueezed tensors - same as output but with size 1
    # in the stack dimension
    unsqueezed_shape = output_shape.copy()
    unsqueezed_shape[dim] = 1

    # Reshape each input tensor to add the new dimension
    unsqueezed_tensors = []
    for tensor in tensors:
        # TODO: remove when concat supports tiled uint32
        tensor = ttnn.reshape(tensor, unsqueezed_shape)
        if tensor.layout == ttnn.TILE_LAYOUT and tensor.dtype == ttnn.uint32:
            tensor = ttnn.to_layout(tensor, ttnn.ROW_MAJOR_LAYOUT)
        unsqueezed_tensors.append(tensor)

    # Concatenate all reshaped tensors along the stack dimension
    return ttnn.concat(unsqueezed_tensors, dim)


@torch.fx.wrap
def all(tensor, num_elements):
    """
    Check if all elements in the tensor are non-zero.

    Args:
        tensor (Tensor): The input tensor to check.
        num_elements (int): The number of elements in the tensor.

    Returns:
        bool: True if all elements in the tensor are non-zero, False otherwise.
    """
    # Check if all elements in the tensor are non-zero
    # by comparing the sum of non-zero elements to the total number of elements
    neq_zero = ttnn.ne(tensor, 0)
    total_none_zero = ttnn.sum(neq_zero)
    return ttnn.eq(total_none_zero, num_elements)


"""
replicate_tensor, shard_tensor, and concat_tensor are needed to propagate shape data throughout the computation graph. These wrappers just replicate the shape change that occurs from the actual ttnn ops without requiring access to a TtnnDevice. It is expected that they are substituted back out during the ToTtPass.
TODO: Find a better way to propagate shapes
"""


@torch.fx.wrap
def replicate_tensor(tensor):
    return tensor


@torch.fx.wrap
def shard_tensor(tensor, dim, num_devices):
    return torch.chunk(tensor, num_devices, dim)[0]


@torch.fx.wrap
def concat_tensor(tensor, dim, num_devices):
    sharded_version = [tensor] * num_devices
    return torch.concat(sharded_version, dim)


# TODO: Support compute kernel config
@torch.fx.wrap
def native_layer_norm(
    input_tensor: ttnn.Tensor,
    in_tensor_shape: torch.Size,
    mean_rstd_shape: torch.Size,
    ttnn_mean_rstd_shape: torch.Size,
    ttnn_dtype: ttnn.DataType,
    norm_dims: int,
    gamma: ttnn.Tensor,
    beta: ttnn.Tensor,
    epsilon: ttnn.Tensor,
    use_mean: bool,
    use_rstd: bool,
    device: ttnn.Device,
):
    if not use_mean and not use_rstd:
        output = ttnn.layer_norm(input_tensor, epsilon=epsilon, weight=gamma, bias=beta)
        return (output, None, None)

    output = ttnn.empty(in_tensor_shape, device=device, layout=ttnn.TILE_LAYOUT, dtype=ttnn_dtype)

    # moreh.layer_norm does not generate correct mean or rstd if the shape keeps the normalized dims
    # https://github.com/tenstorrent/tt-metal/issues/22110
    if use_mean:
        mean = ttnn.empty(ttnn_mean_rstd_shape, device=device, layout=ttnn.TILE_LAYOUT, dtype=ttnn_dtype)
    else:
        mean = None

    if use_rstd:
        rstd = ttnn.empty(ttnn_mean_rstd_shape, device=device, layout=ttnn.TILE_LAYOUT, dtype=ttnn_dtype)
    else:
        rstd = None

    output, mean, rstd = ttnn.operations.moreh.layer_norm(
        input_tensor, norm_dims, epsilon, gamma, beta, output=output, mean=mean, rstd=rstd
    )

    if use_mean:
        mean = ttnn.reshape(mean, mean_rstd_shape)
    if use_rstd:
        rstd = ttnn.reshape(rstd, mean_rstd_shape)

    return (output, mean, rstd)



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import torch
from torch.fx.passes.infra.pass_base import PassBase, PassResult
from torch_ttnn.passes.analysis.input_analysis_pass import PrimalTag

import ttnn


def propagate_sharding_to_users(node, shard_dim, concat_size):
    seen_set = set()

    node_list = [(node, shard_dim, concat_size)]

    while len(node_list) > 0:
        (node, shard_dim, concat_size) = node_list.pop(-1)

        if node.op == "output" or node in seen_set:
            continue

        seen_set.add(node)

        node.meta["is_sharded"] = True

        # permute changes which dimension is sharded
        if node.target == torch.ops.aten.permute.default:
            shard_dim = node.args[1].index(shard_dim)
        node.meta["shard_dim"] = shard_dim

        # if any ops change the size after concatenating all shards, adjust here
        node.meta["concat_size"] = concat_size

        for user in node.users:
            if user not in seen_set:
                node_list.append((user, shard_dim, concat_size))


class MultiDeviceShardAnalysisPass(PassBase):
    def __init__(self, device):
        super().__init__()

        self.device = device

    def call(self, gm: torch.fx.GraphModule):
        if self.device.get_num_devices() < 2:
            modified = False
            return PassResult(gm, modified)

        node_list = list(gm.graph.nodes)
        for node in node_list:
            if node.op == "placeholder" and node.meta.get("primal_tag") == PrimalTag.ARGUMENT:
                shard_dim = 0
                concat_size = node.meta["val"].shape[shard_dim]

                # TODO: remove limitation that tensor size in sharded dimension is divisible by number of devices in mesh
                if concat_size % self.device.get_num_devices() != 0:
                    raise RuntimeError(
                        "Data Parallel runs currently only support sharding dimensions that are multiples of the number of devices"
                    )

                propagate_sharding_to_users(node, shard_dim=shard_dim, concat_size=concat_size)

        modified = False
        return PassResult(gm, modified)



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0

import torch
import ttnn
from torch.fx import Node
from torch_ttnn.passes.patterns.pattern_matcher_base import PatternMatcherBase
from typing import List, Tuple


class SoftMaxPatterns(PatternMatcherBase):
    """Pattern matcher for attention softmax operations."""

    def match_pattern(self) -> List[Tuple[torch.fx.Node, ...]]:
        """
        Match the attention softmax pattern:
        multiply(x) -> add(attention_mask) -> softmax(numeric_stable=True)

        This pattern represents:
        1. Division by sqrt(head_size)
        2. Addition of attention mask
        3. Numerically stable softmax computation

        The expected output will be:
        ttnn.transformer.attention_softmax(input_tensor, head_size, attention_mask)

        Please note that we are doing this in place
        """
        matches = []

        # Find all multiply nodes with something similar to a scale
        # (int, float bigger than 0)
        multiply_nodes = self._find_nodes_of_type(ttnn.multiply)
        for multiply in multiply_nodes:
            # Check for the scale factor (1/sqrt(head_size))
            # this number should be always be positive
            if not (
                len(multiply.args) > 1
                and isinstance(multiply.args[1], (float))
                and multiply.args[1] > 0
                and multiply.args[1] < 1
            ):  # since is 1/sqrt(head_size), it should be less than 1
                continue

            # Find add operation that combines with attention mask
            add = self._find_exclusive_user_of_type(multiply, ttnn.add)
            if not add:
                continue

            # Verify add has both multiply output and attention mask
            if not (multiply in add.args and add.args[0] == multiply):
                continue

            # Find softmax operation with numeric stability
            softmax = self._find_exclusive_user_of_type(add, ttnn.softmax)
            if not softmax:
                continue

            matches.append((multiply, add, softmax))

        return matches

    def replace_pattern(self, matches: List[Tuple[torch.fx.Node, ...]]) -> None:
        for match in matches:
            multiply, add, softmax = match

            with self.gm.graph.inserting_after(softmax):
                # Get the input tensor and attention mask
                input_tensor = multiply.args[0]
                scale = multiply.args[1]
                head_size = None if scale is None else (int(1 / (scale * scale)) if scale < 1 else int(scale))
                attention_mask = add.args[1]
                fused_node = self.gm.graph.call_function(
                    ttnn.transformer.attention_softmax,
                    args=(input_tensor,),
                    kwargs={"head_size": head_size, "attention_mask": attention_mask},
                )

                # Connect output to the next node
                softmax.replace_all_uses_with(fused_node)

            # Remove old nodes in reverse order
            nodes_to_remove = [softmax, add, multiply]
            self.safe_remove_nodes(nodes_to_remove)

        self.gm.graph.lint()
        self.gm.recompile()
        return matches



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import torch
from torch.fx.passes.infra.pass_base import PassBase, PassResult
from enum import Enum


class PrimalTag(Enum):
    """Enumeration of compiled model inputs.

    It is expected that PARAMETER and BUFFER tensors do not change between inference runs, but ARGUMENT tensors may change.

    :param PARAMETER: Tensor tracked by optimizer during training. Represents weights and biases.
    :param BUFFER: Tensor not updated during training but still part of model state. Normally used for non-trainable data like fixed weights or running statistics in batch norm.
    :param ARGUMENT: Tensors not tracked by model and not persistent between calls. Can be input data or configuration options to module's methods.
    """

    PARAMETER = 1
    BUFFER = 2
    ARGUMENT = 3


class InputAnalysisPass(PassBase):
    def __init__(self, num_parameters, num_buffers, num_arguments):
        super().__init__()

        self.num_parameters = num_parameters
        self.num_buffers = num_buffers
        self.num_arguments = num_arguments

    def call(self, gm: torch.fx.GraphModule):
        """Marks inputs for the compiled function as parameters, buffers, or arguments.

        This relies on how PyTorch behaves under the hood. Namely, PyTorch will lay out function arguments in the order of parameters, then buffers, then arguments. We can rely on this layout order and the number of parameters, buffers, and arguments passed in from AOTAutograd to our backend function.

        :param gm: Graph module for the function being compiled.
        :return: Pass result with the updated graph module and whether the graph was modified.
        :rtype: PassResult[torch.fx.GraphModule, bool]
        """

        def annotate_node(node, tag):
            node.meta["primal_tag"] = tag

        modified = False
        placeholder_counter = 0

        node_list = list(gm.graph.nodes)
        for node in node_list:
            if node.op == "placeholder":
                if placeholder_counter < self.num_parameters:
                    annotate_node(node, PrimalTag.PARAMETER)
                elif placeholder_counter < self.num_parameters + self.num_buffers:
                    annotate_node(node, PrimalTag.BUFFER)
                else:
                    annotate_node(node, PrimalTag.ARGUMENT)
                placeholder_counter += 1

        return PassResult(gm, modified)



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import torch
import ttnn


def is_getitem(n: torch.fx.node.Node):
    if n.op == "call_function" and n.target.__name__ == "getitem":
        return True
    return False


def group_norm_rep(gm: torch.fx.GraphModule):
    changed = False
    for n in gm.graph.nodes:
        if not (
            isinstance(n, torch.fx.node.Node)
            and n.op == "call_function"
            and n.target == torch.ops.aten.native_group_norm.default
            and len(n.users) == 1
        ):
            continue

        next_n = next(iter(n.users))
        if not (isinstance(next_n, torch.fx.node.Node) and next_n.op == "call_function" and is_getitem(next_n)):
            continue

        n_kwargs = {
            "input": n.args[0],
            "weight": n.args[1],
            "bias": n.args[2],
            "N": n.args[3],
            "C": n.args[4],
            "HxW": n.args[5],
            "group": n.args[6],
            "eps": n.args[7],
        }
        ttnn_kwargs = {
            "num_groups": n_kwargs["group"],
            "epsilon": n_kwargs["eps"],
        }
        if n_kwargs["weight"] is not None:
            ttnn_kwargs["weight"] = n_kwargs["weight"]
        if n_kwargs["bias"] is not None:
            ttnn_kwargs["bias"] = n_kwargs["bias"]

        g = gm.graph
        with g.inserting_after(next_n):
            ttnn_node = g.call_function(ttnn.group_norm, (n_kwargs["input"],), ttnn_kwargs)
            next_n.replace_all_uses_with(ttnn_node)
        changed = True

    if changed:
        gm.graph.eliminate_dead_code()
    return gm


def layer_norm_rep(gm: torch.fx.GraphModule):
    changed = False
    for n in gm.graph.nodes:
        if not (
            isinstance(n, torch.fx.node.Node)
            and n.op == "call_function"
            and n.target == torch.ops.aten.native_layer_norm.default
            and len(n.users) == 1
        ):
            continue

        next_n = next(iter(n.users))
        if not (isinstance(next_n, torch.fx.node.Node) and next_n.op == "call_function" and is_getitem(next_n)):
            continue

        n_kwargs = {
            "input": n.args[0],
            "normalized_shape": n.args[1],
            "weight": n.args[2],
            "bias": n.args[3],
            "eps": n.args[4],
        }
        ttnn_kwargs = {"normalized_shape": n_kwargs["normalized_shape"]}
        if n_kwargs["weight"] is not None:
            ttnn_kwargs["weight"] = n_kwargs["weight"]
        if n_kwargs["bias"] is not None:
            ttnn_kwargs["bias"] = n_kwargs["bias"]

        g = gm.graph
        with g.inserting_after(next_n):
            ttnn_node = g.call_function(ttnn.layer_norm, (n_kwargs["input"],), ttnn_kwargs)
            next_n.replace_all_uses_with(ttnn_node)
        changed = True

    if changed:
        gm.graph.eliminate_dead_code()
    return gm


customized_rep_list = [
    group_norm_rep,
    # layer_norm_rep
]



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import torch
import ttnn
from torch.fx.passes.infra.pass_base import PassBase, PassResult


def _eliminate_pair(node, func, pre_func):
    """
    Eliminate redundant pattern of paired data movement, such as from_device => to_device.
    """
    changed = False
    available_func_list = [
        ttnn.from_device,
        ttnn.to_device,
        ttnn.from_torch,
        ttnn.to_torch,
    ]
    if func not in available_func_list or pre_func not in available_func_list:
        return changed
    if len(node.users) == 0:
        return changed
    if node.op != "call_function" or node.target != func:
        return changed
    pre_node = node.args[0]
    if type(pre_node) != torch.fx.node.Node:
        return changed
    if pre_node.op != "call_function" or pre_node.target != pre_func:
        return changed
    node.replace_all_uses_with(pre_node.args[0])
    changed = True
    return changed


class EliminateCoreopsPass(PassBase):
    def call(self, gm: torch.fx.GraphModule):
        modified = False
        for node in gm.graph.nodes:
            modified |= _eliminate_pair(node, ttnn.to_device, ttnn.from_device)
            modified |= _eliminate_pair(node, ttnn.from_device, ttnn.to_device)
            modified |= _eliminate_pair(node, ttnn.to_torch, ttnn.from_torch)
            modified |= _eliminate_pair(node, ttnn.from_torch, ttnn.to_torch)
        if modified:
            gm.graph.eliminate_dead_code()
        return PassResult(gm, modified)



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import torch
from typing import List
from torch_ttnn.utils import GraphCleanup

"""
AOT Autograd has an optimization where if it determines that the storage of the
output is the same as the input, it will return data from the input. The output
is checked if it's an alias of the input. This becomes problematic when we
have data transfer between host and device. The issue has been raised before.
See: https://github.com/pytorch/pytorch/issues/108079

One workaround is to insert a clone op after every input node before the graph
is completely lowered to aten ops. This can be accomplished by pointing the
entry function with a `torch.fx.GraphModule` parameter to the `backend`
positional argument for `torch.compile`. At this point, the graph will have
higher-level torch ops where clone ops can be inserted. Then define another
function that will be passed to the `fw_compiler/bw_compiler` parameter for
`aot_autograd`. This is when the input aliasing metadata will be determined
and the graph will be lowered to aten ops. After this, the clone ops can
be removed with another pass.

The method is inspired by TensorRT:
https://github.com/pytorch/TensorRT/commit/7daa1120dc1bc72d6f92f1e7aa2b357a65b6ea08
"""


# Insert aten.clone nodes after every input to prevent input aliasing
def insert_clones_for_input_aliasing(gm: torch.fx.GraphModule) -> torch.fx.GraphModule:
    # Helper function to determine if a node has in-place mutation. For example, the torch op
    # `self_tensor.copy_(source_tensor)` will replace the self_tensor with the source_tensor.
    # The arguments for this op will be (self_tensor, source_tensor) in this order.
    # TODO: Is there a better way to determine if an op mutates an input without listing manually?
    def is_node_mutated(node):
        torch_ops_that_mutate = ["copy_", "torch.slice_scatter"]
        # the mutated tensor will always be the first argument, args[0]
        return any(
            [
                str(user.target) in torch_ops_that_mutate and len(user.args) > 1 and node == user.args[0]
                for user in list(node.users.keys())
            ]
        )

    # get input tensor nodes only, but skip if the input is also mutated
    input_nodes = [
        node
        for node in gm.graph.nodes
        if (node.op == "placeholder" and node.meta["grapharg"].is_tensor and not is_node_mutated(node))
    ]

    modified = False
    for node in input_nodes:
        """TODO(kevinwuTT): This does not work if inserting right after the node itself.
        Only works if inserting after all of the input_nodes.
        TypeError: forward() missing `n` required positional arguments
        Somehow the argument list will get truncated.
        """
        with gm.graph.inserting_after(input_nodes[-1]):
            clone_node = gm.graph.call_function(torch.ops.aten.clone.default, args=(node,))
            node.replace_all_uses_with(
                clone_node,
                delete_user_cb=lambda node: node != clone_node,
            )
            modified = True

    if modified:
        """
        Do not call `eliminate_dead_code()` before `aot_autograd` because the function has not
        been functionalized yet. This means some other mutations will be eliminate
        inadvertently.

        Example:
        ```
        1: def f():
        2:    mask = torch.full((1, 1), 1)
        3:    mask_cond = torch.full((1, 1), True)
        4:    masked_fill = mask.masked_fill_(mask_cond, 0)
        5:    mask_1 = mask.to(torch.bfloat16)
        6:    return mask_1
        ```

        In the function above, `eliminate_dead_code()` does not know that `mask.masked_fill_` mutates
        the `mask` tensor in-place. Therefore, it sees that `masked_fill` is not used and will eliminate lines 3 and 4. The result of the function above should be `tensor([[0.]], dtype=torch.bfloat16)`. However, after code elimination, the function will become below and
        the result will be `tensor([[1.]], dtype=torch.bfloat16)` which is incorrect.

        ```
        1: def f():
        2:    mask = torch.full((1, 1), 1)
        3:    mask_1 = mask.to(torch.bfloat16)
        4:    return mask_1
        ```
        """
        gm.graph.lint()
        gm.recompile()

    return gm


def remove_clones_for_input_aliasing(gm: torch.fx.GraphModule) -> torch.fx.GraphModule:
    """Remove the clone ops inserted to handle input aliasing
    opcode         name             target               args
    -------------  ---------------  -------------------  ------------
    placeholder    l_input_         L_input_             ()
    call_function  clone_default    aten.clone.default   (l_input_,)
    call_function  op               <op_func>            (clone_default)
    """
    modified = False
    for node in gm.graph.nodes:
        if (
            node.op == "placeholder"
            and len(node.users) == 1
            and list(node.users)[0].target == torch.ops.aten.clone.default
        ):
            clone_node = list(node.users)[0]
            clone_node.replace_all_uses_with(node)
            gm.graph.erase_node(clone_node)

            modified = True

    if modified:
        gm = GraphCleanup(gm)

    return gm



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import ttnn
import torch
import torch._dynamo
from typing import Optional, Tuple, List


class PatternMatcherBase:
    """
    Base class for pattern matching in FX graphs.

    All pattern matchers should return a list of tuples of torch.fx.Node objects.
    """

    def __init__(self, gm: torch.fx.GraphModule):
        self.gm = gm

    def _is_ttnn_op(self, node: torch.fx.Node, op_type) -> bool:
        """Helper to check if a node is a specific ttnn operation."""
        return node.op == "call_function" and node.target == op_type

    def _find_exclusive_user_of_type(self, node: torch.fx.Node, op_type) -> Optional[torch.fx.Node]:
        """
        Find a node that is exclusively consumed by a single user of the specified type.
        Also ensures the user is not ttnn.to_torch.

        Args:
            node: The node to check users for
            op_type: The operation type to match against

        Returns:
            The single user of specified type if found and exclusive, None otherwise
        """
        users = list(node.users)
        if len(users) != 1:  # Must have exactly one user
            return None

        user = users[0]
        if self._is_ttnn_op(user, ttnn.to_torch):  # Check not ttnn.to_torch
            return None

        return user if self._is_ttnn_op(user, op_type) else None

    def _find_single_user_of_type(self, node: torch.fx.Node, op_type) -> Optional[torch.fx.Node]:
        """Find a single user of the given node with the specified operation type,
        ensuring no immediate user is a ttnn.to_torch operation."""

        # Why not ttnn.to_torch? Simply because it breaks the tree structure of the graph.
        # if I do a replacement, let's say linear and then we have a ttnn.to_torch with one of
        # the intermediate nodes, I need to do something fancier to avoid breaking the graph.
        # I prefer to be conservative when reducing nodes instead of adding an extra layer
        # of complexity to the pattern matcher.
        if any(self._is_ttnn_op(user, ttnn.to_torch) for user in node.users):
            return None

        matching_users = [user for user in node.users if self._is_ttnn_op(user, op_type)]
        return matching_users[0] if len(matching_users) == 1 else None

    def _is_node_in_args(self, target_node: torch.fx.Node, source_node: torch.fx.Node) -> bool:
        """Check if target_node is one of the arguments of source_node."""
        return target_node in source_node.args

    def _find_nodes_of_type(self, op_type) -> List[torch.fx.Node]:
        """Find all nodes of a specific operation type in the graph."""
        return [node for node in self.gm.graph.nodes if self._is_ttnn_op(node, op_type)]

    def match_pattern(self) -> List[Tuple[torch.fx.Node, ...]]:
        """
        Abstract method to be implemented by subclasses.
        Should return a list of matched patterns as tuples of torch.fx.Node.
        """
        raise NotImplementedError("Subclasses must implement match_pattern")

    def replace_pattern(self, matches: List[Tuple[torch.fx.Node, ...]]) -> None:
        """
        Abstract method to be implemented by subclasses.
        Should implement the pattern replacement logic.
        """
        raise NotImplementedError("Subclasses must implement replace_pattern")

    def safe_remove_nodes(self, nodes: List[Optional[torch.fx.Node]]) -> None:
        """
        Remove nodes from the graph. Since ttnn.to_torch checks are handled during pattern matching,
        we can safely remove these nodes.

        Args:
            nodes: List of nodes to be removed (can contain None values which will be skipped)
        """

        for node in nodes:
            if node is not None:
                self.gm.graph.erase_node(node)



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import logging
import torch
import ttnn
from torch_ttnn.passes.analysis.input_analysis_pass import PrimalTag
from torch_ttnn.utils import (
    GraphCleanup,
    TtnnRowMajorLayout,
    TtnnTileLayout,
    TtnnDevice,
    TtnnBfloat16,
    TtnnUint32,
    HasValidPageSize,
    get_dtype,
    get_meta_val_attr,
)
from dataclasses import dataclass
from enum import Enum
from typing import Union, Type, Literal
from operator import getitem

from torch.fx.passes.infra.pass_base import PassBase, PassResult
from . import target_wrappers


def is_function_call(node) -> bool:
    if not isinstance(node, torch.fx.node.Node):
        return False
    return node.op == "call_function"


TTNN_POINTWISE_UNARY_OPS = [
    ttnn.abs,
    ttnn.acos,
    ttnn.acosh,
    ttnn.asin,
    ttnn.asinh,
    ttnn.atan,
    ttnn.atanh,
    ttnn.ceil,
    ttnn.cos,
    ttnn.cosh,
    ttnn.elu,
    ttnn.erf,
    ttnn.exp,
    ttnn.expm1,
    ttnn.floor,
    ttnn.gelu,
    ttnn.hardsigmoid,
    ttnn.hardswish,
    ttnn.hardtanh,
    ttnn.isinf,
    ttnn.isnan,
    ttnn.leaky_relu,
    ttnn.log,
    ttnn.log10,
    ttnn.log1p,
    ttnn.log2,
    ttnn.logical_not,
    ttnn.min,
    ttnn.neg,
    ttnn.reciprocal,
    ttnn.relu,
    ttnn.remainder,
    ttnn.round,
    ttnn.rsqrt,
    ttnn.sigmoid,
    ttnn.softmax,
    ttnn.sign,
    ttnn.sin,
    ttnn.sinh,
    ttnn.silu,
    ttnn.sqrt,
    ttnn.tan,
    ttnn.tanh,
    ttnn.trunc,
    ttnn.bitwise_not,
]


TTNN_POINTWISE_BINARY_OPS = [
    ttnn.add,
    ttnn.atan2,
    ttnn.div,
    ttnn.eqz,
    ttnn.floor_div,
    ttnn.gez,
    ttnn.ge,
    ttnn.gtz,
    ttnn.isclose,
    ttnn.ldexp,
    ttnn.lez,
    ttnn.logaddexp,
    ttnn.logaddexp2,
    ttnn.le,
    ttnn.ltz,
    ttnn.nextafter,
    ttnn.nez,
    ttnn.polyval,
    ttnn.squared_difference,
    ttnn.eq,
    ttnn.gt,
    ttnn.logical_and,
    ttnn.logical_or,
    ttnn.logical_xor,
    ttnn.lt,
    ttnn.maximum,
    ttnn.minimum,
    ttnn.mul,
    ttnn.ne,
    ttnn.pow,
    ttnn.rsub,
    ttnn.sub,
    ttnn.xlogy,
    # ttnn.add_and_apply_activation,  # ttnn has no add_and_apply_activation, remote the comment in the future when it has
    # ttnn.add_and_apply_activation_,  # ttnn has no add_and_apply_activation, remote the comment in the future when it has
]

TTNN_POINTWISE_TRINARY_OPS = [
    ttnn.addcdiv,
    ttnn.addcmul,
    ttnn.mac,
    ttnn.where,
]

TTNN_MATRIX_MULTIPLICATION_OPS = [
    ttnn.matmul,
    ttnn.linear,
]

TTNN_DATAMOVE_OPS = [
    ttnn.concat,
    ttnn.pad,
    ttnn.permute,
    ttnn.reshape,
    ttnn.experimental.view,
    ttnn.sharded_to_interleaved,
    ttnn.slice,
    ttnn.split,
    ttnn.squeeze,
    ttnn.to_layout,
    ttnn.transpose,
    ttnn.unsqueeze,
]

TTNN_TARGET_WRAPPERS = [
    target_wrappers.clone,
    target_wrappers.repeat,
    target_wrappers.pack_to_tuple,
    target_wrappers.move_to_host,
    target_wrappers.conv,
    target_wrappers.roll,
    target_wrappers.stack,
    target_wrappers.all,
    target_wrappers.concat_tensor,
    target_wrappers.native_layer_norm,
]

TTNN_NORM_OPS = [
    ttnn.group_norm,
]

TTNN_POOL_OPS = [
    ttnn.max_pool2d,
    ttnn.avg_pool2d,
]

TTNN_ROW_LAYOUT_OPS = set(
    [
        ttnn.slice,
        target_wrappers.roll,
        ttnn.argmax,
    ]
)

# Operations that might output row major layouts based on ToTtPass implementation
TTNN_MAYBE_ROW_OPS = set(
    [
        ttnn.pad,
        ttnn.concat,
    ]
)

TTNN_TRANSFORMER_OPS = [
    ttnn.transformer.scaled_dot_product_attention,
    ttnn.transformer.attention_softmax,
    ttnn.transformer.split_query_key_value_and_split_heads,
    ttnn.transformer.concatenate_heads,
]

TTNN_HOST_ONLY_OPS = set()


# For operations limitations
# See https://github.com/tenstorrent-metal/tt-metal/blob/main/ttnn/README.md?plain=1#L19
def is_tt_compute(node) -> bool:
    if not is_function_call(node):
        return False

    # if node is the built-in function "getitem", the result of split
    # we have to check the input of split
    if node.op == "call_function" and node.target.__name__ == "getitem":
        return is_tt_compute(node.args[0])

    return node.target in set(
        TTNN_POINTWISE_UNARY_OPS
        + TTNN_POINTWISE_BINARY_OPS
        + TTNN_POINTWISE_TRINARY_OPS
        + TTNN_MATRIX_MULTIPLICATION_OPS
        + TTNN_TARGET_WRAPPERS
        + TTNN_DATAMOVE_OPS
        + TTNN_NORM_OPS
        + TTNN_POOL_OPS
        + TTNN_TRANSFORMER_OPS
        + [
            ttnn.embedding,
            ttnn.ones,
            ttnn.tril,
            ttnn.arange,
            ttnn.zeros,
            ttnn.zeros_like,
            ttnn.mean,
            ttnn.moreh_cumsum,
            ttnn.clip,
            ttnn.squeeze,
            ttnn.full,
            ttnn.as_tensor,
            ttnn.expand,
            ttnn.moreh_cumsum,
            ttnn.sum,
            ttnn.typecast,
            ttnn.argmax,
            ttnn.fill,
            ttnn.empty,
        ]
    )


def is_tt_data_move(node) -> bool:
    if not is_function_call(node):
        return False
    return node.target in [
        ttnn.from_device,
        ttnn.to_device,
        ttnn.from_torch,
        ttnn.to_torch,
        ttnn.to_layout,
        ttnn.MemoryConfig,
        ttnn.Shape,
    ]


def is_tt(node):
    return is_tt_compute(node) or is_tt_data_move(node)


def is_torch_to_ttnn(src_node, dst_node) -> bool:
    if isinstance(src_node, (int, float, list, tuple)) or not isinstance(src_node, torch.fx.node.Node):
        return False
    return is_tt_compute(dst_node) and not is_tt(src_node) and not (dst_node.target == ttnn.as_tensor)


def is_ttnn_to_torch(src_node, dst_node) -> bool:
    return is_tt_compute(src_node) and not is_tt(dst_node)


def is_ttnn_to_ttnn(src_node, dst_node):
    if not is_tt(src_node) or not is_tt(dst_node):
        return False
    return True


class NodeInputAligner:
    def __init__(self, graph, device):
        self.graph = graph
        self.device = device
        # aligned_node_dict maps DataMoveSpec to aligned version of the node to prevent calling the same data movement twice
        self.aligned_node_dict = {}

        # fields for data parallel
        self.shard_to_mesh = dict()
        self.replicate_to_mesh = None
        self.concat_to_mesh = dict()

    class InputSiteType(Enum):
        ARGS = 1
        KWARGS = 2
        ARGS_TUPLE = 3
        KWARGS_TUPLE = 4

    @dataclass(unsafe_hash=True)
    class AlignSpecFromTorch:
        input_node: torch.fx.node.Node
        device: Union[None, Type[TtnnDevice], Literal["host"]]
        layout: Union[None, Type[TtnnTileLayout], Type[TtnnRowMajorLayout]]
        dtype: Union[None, Type[TtnnBfloat16], Type[TtnnUint32]]

    @dataclass(unsafe_hash=True)
    class AlignSpecToTorch:
        input_node: torch.fx.node.Node
        dtype: Union[None, Type[TtnnBfloat16], Type[TtnnUint32], Literal["by_node_meta"]]

    @dataclass(unsafe_hash=True)
    class AlignSpecInTtnn:
        input_node: torch.fx.node.Node
        device: Union[None, Type[TtnnDevice], Literal["host"]]
        layout: Union[None, Type[TtnnTileLayout], Type[TtnnRowMajorLayout]]
        dtype: Union[None, Type[TtnnBfloat16], Type[TtnnUint32]]

    def _reset_to_default_layout(self, input_node, spec):
        # split(list of tensor with row major layout) => getitem(row major layout)
        # convert back to tile layout
        if input_node.target == getitem and input_node.args[0].target == ttnn.split:
            spec.layout = TtnnTileLayout

        # re-tilize max_pool2d after sharded_to_interleaved call - may be able to remove after #418
        if input_node.target == ttnn.sharded_to_interleaved and input_node.args[0].target in [
            ttnn.max_pool2d,
            ttnn.avg_pool2d,
        ]:
            spec.layout = TtnnTileLayout

        # be overly cautious and convert to tile layout. These could already be tilized
        # TODO: only insert layout if needed
        if input_node.target in [ttnn.ones, ttnn.ones_like, ttnn.zeros, ttnn.zeros_like]:
            spec.layout = TtnnTileLayout

        # legalize to the default layout and device
        if input_node.target in TTNN_ROW_LAYOUT_OPS:
            spec.layout = TtnnTileLayout
        if input_node.target in TTNN_MAYBE_ROW_OPS:
            # for now, convert to tile (might be nop)
            # TODO: only insert to_layout call if needed
            spec.layout = TtnnTileLayout
        if input_node.target in TTNN_HOST_ONLY_OPS:
            spec.device = TtnnDevice

        return spec

    def _align_special_cases(self, node, spec, input_site, input_site_type: InputSiteType):
        if node.target == ttnn.embedding:
            # Embedding is not as accurate with TileLayout (allclose with torch.embedding fails)
            if input_site == 0:
                # First input must be uint32
                spec.dtype = TtnnUint32
                spec.layout = TtnnRowMajorLayout
        if node.target in [ttnn.slice, target_wrappers.roll] and (
            input_site_type == self.InputSiteType.ARGS and input_site == 0
        ):
            # Slice can only unpad tilized tensors with full tiles. Make input row major to be safe for now
            # Roll is included because it calls slice under the hood
            spec.layout = TtnnRowMajorLayout
        if node.target == ttnn.split and input_site == 0:
            # Runtime error for tilized inputs to split
            spec.layout = TtnnRowMajorLayout
        if node.target == ttnn.argmax and (input_site_type == self.InputSiteType.ARGS and input_site == 0):
            # According to documentation, argmax input must be BFLOAT16 and ROW_MAJOR
            spec.dtype = TtnnBfloat16
            spec.layout = TtnnRowMajorLayout
        if (
            node.target == target_wrappers.stack
            and isinstance(spec, self.AlignSpecFromTorch)
            and get_dtype(node) in [torch.int32, torch.int64]
        ):
            # This allows ViLT to work by coercing stack inputs to be uint32
            # TODO: remove this and handle stack inputs more generally
            spec.dtype = TtnnUint32
        if node.target == target_wrappers.conv and input_site == 1:
            # TODO(#417, tt-metal#15893): weight currently needs to be on host and can't be moved to device first
            spec.layout = TtnnRowMajorLayout
            spec.device = "host"
        if (
            node.target == ttnn.subtract
            and isinstance(spec, self.AlignSpecFromTorch)
            and get_dtype(node) in [torch.int32, torch.int64]
        ):
            # Needed to have GPT2 pass: from_torch -> subtract -> remainder -> indexing but subtract may go below 0
            # TODO: remove when we have signed int support for required ops
            spec.dtype = TtnnBfloat16

        return spec

    def _get_align_spec(self, node, input_node, input_site, input_site_type: InputSiteType):
        if is_torch_to_ttnn(input_node, node):
            # default set these layout for torch to ttnn
            spec_dtype = TtnnBfloat16
            if get_dtype(input_node) in [torch.int32, torch.int64]:
                spec_dtype = TtnnUint32
            spec = self.AlignSpecFromTorch(input_node, TtnnDevice, TtnnTileLayout, spec_dtype)
            spec = self._align_special_cases(node, spec, input_site, input_site_type)
            return spec
        elif is_ttnn_to_torch(input_node, node):
            spec = self.AlignSpecToTorch(input_node, get_dtype(input_node))
            return spec
        elif is_ttnn_to_ttnn(input_node, node):
            # default do nothing between ttnn to ttnn
            spec = self.AlignSpecInTtnn(input_node, None, None, None)
            spec = self._reset_to_default_layout(input_node, spec)
            spec = self._align_special_cases(node, spec, input_site, input_site_type)

            if spec.device is None and spec.layout is None and spec.dtype is None:
                return None
            return spec
        else:
            return None

    def _change_layout(self, spec):
        need_from_device = spec.device == "host"
        need_to_layout = spec.layout is not None
        need_to_device = spec.device == TtnnDevice

        input_node = spec.input_node

        # mesh device tensors need layout on device
        if need_to_device:
            input_node = self.graph.call_function(ttnn.to_device, (input_node,), {"device": TtnnDevice()})

        if need_to_layout:
            input_node = self.graph.call_function(ttnn.to_layout, (input_node, spec.layout()))

        if need_from_device:
            input_node = self.graph.call_function(ttnn.from_device, (input_node,))

        return input_node

    def _extract_args_kwargs_from_spec(self, spec):
        if isinstance(spec, self.AlignSpecFromTorch):
            kwargs = {}
            args = (spec.input_node,)
            if spec.device is not None and spec.device != "host":
                kwargs["device"] = spec.device()
            if spec.layout is not None:
                kwargs["layout"] = spec.layout()
            if spec.dtype is not None:
                kwargs["dtype"] = spec.dtype()

            # A bit tricky: here we actually end up replacing the input_node with its equivalent from_torch call by accessing its arguments. Sharding and replicating are determined by the wrapper name
            if self.device.get_num_devices() > 1:
                if spec.input_node.target == target_wrappers.shard_tensor:
                    actual_inp_node, shard_dim, _ = spec.input_node.args

                    if self.shard_to_mesh.get(shard_dim) is None:
                        self.shard_to_mesh[shard_dim] = self.graph.call_function(
                            ttnn.ShardTensorToMesh, args=(TtnnDevice(),), kwargs={"dim": shard_dim}
                        )

                    mesh_mapper = self.shard_to_mesh[shard_dim]
                    args = (actual_inp_node,)
                else:
                    if self.replicate_to_mesh is None:
                        self.replicate_to_mesh = self.graph.call_function(
                            ttnn.ReplicateTensorToMesh, args=(TtnnDevice(),)
                        )

                    mesh_mapper = self.replicate_to_mesh
                    args = spec.input_node.args
                kwargs["mesh_mapper"] = mesh_mapper

            return args, kwargs

        elif isinstance(spec, self.AlignSpecToTorch):
            kwargs = {"dtype": spec.dtype}
            args = (spec.input_node,)

            # A bit tricky: here we actually end up replacing the input_node with its equivalent to_torch call by accessing its arguments
            if self.device.get_num_devices() > 1:
                if spec.input_node.target == target_wrappers.concat_tensor:
                    actual_inp_node, shard_dim, _ = spec.input_node.args

                    if self.concat_to_mesh.get(shard_dim) is None:
                        self.concat_to_mesh[shard_dim] = self.graph.call_function(
                            ttnn.ConcatMeshToTensor, args=(TtnnDevice(),), kwargs={"dim": shard_dim}
                        )

                    composer = self.concat_to_mesh[shard_dim]
                    kwargs["mesh_composer"] = composer
                    args = (actual_inp_node,)

            return args, kwargs

        elif isinstance(spec, self.AlignSpecInTtnn):
            return (), {}

        else:
            raise RuntimeError(f"Cannot create aligned node for unknown spec ({spec})")

    def _create_aligned_node(self, spec, node, input_site):
        args, kwargs = self._extract_args_kwargs_from_spec(spec)

        if isinstance(spec, self.AlignSpecFromTorch):
            # TODO: Add mesh support for native integration
            if (
                spec.input_node.op == "placeholder"
                and (native_device := get_meta_val_attr(spec.input_node, "device"))
                and str(native_device) == "ttnn:0"
            ):
                from torch_ttnn.cpp_extension import ttnn_module

                aligning_nodes = []
                aligning_nodes.append(self.graph.call_function(ttnn_module.get_ttnn_tensor, args, {}))
                # Native device tensors are currently initialized in row-major layout and on device
                # Convert them if necessary
                if (layout := kwargs.get("layout", None)) is not None and isinstance(layout, TtnnTileLayout):
                    aligning_nodes.append(self.graph.call_function(ttnn.to_layout, (aligning_nodes[-1], layout)))
                if "device" not in kwargs:
                    aligning_nodes.append(self.graph.call_function(ttnn.from_device, (aligning_nodes[-1],)))
                return aligning_nodes[-1]
            else:
                return self.graph.call_function(ttnn.from_torch, args, kwargs)

        elif isinstance(spec, self.AlignSpecToTorch):
            return self.graph.call_function(ttnn.to_torch, args, kwargs)

        elif isinstance(spec, self.AlignSpecInTtnn):
            return self._change_layout(spec)

        else:
            raise RuntimeError(f"Cannot create aligned node for unknown spec ({spec})")

    def _connect_aligned_node(self, node, aligned_node, input_site, input_site_type: InputSiteType):
        if input_site_type == self.InputSiteType.ARGS:
            input_idx = input_site
            node.update_arg(input_idx, aligned_node)
        elif input_site_type == self.InputSiteType.KWARGS:
            key = input_site
            node.update_kwarg(key, aligned_node)
        elif input_site_type == self.InputSiteType.ARGS_TUPLE:
            input_idx = input_site[0]
            tuple_idx = input_site[1]
            old_arg = node.args[input_idx]
            new_arg = list(old_arg)
            new_arg[tuple_idx] = aligned_node
            node.update_arg(input_idx, tuple(new_arg))
        elif input_site_type == self.InputSiteType.KWARGS_TUPLE:
            key = input_site[0]
            tuple_idx = input_site[1]
            old_arg = node.kwargs[key]
            new_arg = list(old_arg)
            new_arg[tuple_idx] = aligned_node
            node.update_kwarg(key, tuple(new_arg))

    def align(self, node, input_node, input_site, input_site_type: InputSiteType, first_node):
        # assert input_site_type in ["args", "kwargs", "args_tuple", "kwargs_tuple"]
        data_move_spec = self._get_align_spec(node, input_node, input_site, input_site_type)
        if data_move_spec is None:
            # No need to align input_node
            return 0

        if data_move_spec in self.aligned_node_dict:
            aligned_node = self.aligned_node_dict[data_move_spec]
        else:
            with self.graph.inserting_before(node):
                aligned_node = self._create_aligned_node(data_move_spec, node, input_site)
            self.aligned_node_dict[data_move_spec] = aligned_node

        self._connect_aligned_node(node, aligned_node, input_site, input_site_type)

        return 1


class AddDataMovePass(PassBase):
    """Pass that adds instructions to move data between host and device and align tensor dtype and layout.

    This pass is multi-device aware, since it must marshal tensors differently in the multi-device case. This pass attempts to introduce the minimum amount of operations while maintaining correctness and performance.

    :param device: The device on which a workload will run (either a MeshDevice or Device).
    """

    def __init__(self, device):
        self.device = device

    def call(self, gm: torch.fx.GraphModule):
        SiteType = NodeInputAligner.InputSiteType

        modifications_count = 0
        node_input_aligner = NodeInputAligner(gm.graph, self.device)
        nodes = list(gm.graph.nodes)

        first_node = [node for node in nodes if node.op != "placeholder"][0]

        for node in nodes:
            args = node.args
            for idx, arg in enumerate(args):
                if isinstance(arg, (tuple, list, torch.fx.immutable_collections.immutable_list)):
                    for tuple_idx, tuple_arg in enumerate(arg):
                        modifications_count += node_input_aligner.align(
                            node, tuple_arg, [idx, tuple_idx], SiteType.ARGS_TUPLE, first_node
                        )
                else:
                    modifications_count += node_input_aligner.align(node, arg, idx, SiteType.ARGS, first_node)

            kwargs = node.kwargs
            for key, arg in kwargs.items():
                if isinstance(arg, (tuple, list, torch.fx.immutable_collections.immutable_list)):
                    for tuple_idx, tuple_arg in enumerate(arg):
                        modifications_count += node_input_aligner.align(
                            node, tuple_arg, [key, tuple_idx], SiteType.KWARGS_TUPLE, first_node
                        )
                else:
                    modifications_count += node_input_aligner.align(node, arg, key, SiteType.KWARGS, first_node)

        modified = modifications_count > 0
        GraphCleanup(gm)
        return PassResult(gm, modified)



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import torch
import re


def GraphCleanup(gm: torch.fx.GraphModule) -> torch.fx.GraphModule:
    gm.graph.eliminate_dead_code()
    gm.graph.lint()
    gm.recompile()

    return gm


def get_shape(gm: torch.fx.GraphModule, node_or_shape):
    """
    Get the shape of a node or shape itself.

    Args:
        gm (torch.fx.GraphModule): The GraphModule containing the node.
        node_or_shape: The node or shape to get the shape of. Can be an int, float, torch.Size, list, tuple, torch.fx.node.Node, or torch.fx.proxy.Proxy.

    Returns:
        torch.Size or None: The shape of the node or shape itself, or None if it cannot be determined.
    """
    if isinstance(node_or_shape, torch.fx.proxy.Proxy):
        node_or_shape = node_or_shape.node

    if isinstance(node_or_shape, (int, float)):
        return torch.Size()
    if isinstance(node_or_shape, (torch.Size, list, tuple)):
        return node_or_shape
    if isinstance(node_or_shape, torch.fx.node.Node):
        if (val := node_or_shape.meta.get("val", None)) is not None:
            return val.size()

        if node_or_shape.op == "get_attr":
            if gm is None:
                return None
            val = getattr(gm, node_or_shape.target)
            if isinstance(val, torch.Tensor):
                return val.size()
            if isinstance(val, (int, float)):
                return torch.Size()

    return None


def get_arg(node, index, name, default=None):
    if hasattr(node, "args") and len(node.args) > index:
        return node.args[index]
    if hasattr(node, "kwargs") and name in node.kwargs:
        return node.kwargs[name]
    return default


def get_dtype(node):
    if isinstance(node, torch.fx.node.Node):
        if (val := node.meta.get("val", None)) is not None:
            return val.dtype
    return None


def get_opname(node):
    if str(node.target).startswith("aten."):
        return str(node.target)
    elif hasattr(node.target, "__name__"):
        return node.target.__name__
    elif isinstance(node.op, str):
        return node.target
    else:
        raise


def users_have_getitem(node):
    for user in list(node.users.keys()):
        if user.op != "output" and user.op != "placeholder" and user.target.__name__ == "getitem":
            return user
    return None


def is_operation(node):
    return node.op not in ["placeholder", "output"]


def get_meta_val_attr(node, attr: str):
    """
    Example usage in condition statement:
    # Important to wrap var assignment in parenthesis because of order of operations
    if (var := get_meta_val_attr(node, "attr")) and str(attr) == "some_val":
        ...
    """
    if "val" in node.meta:
        return getattr(node.meta["val"], attr, False)
    else:
        return False


# Certain ops don't support certain shapes and will emit a valid_page_size error
# RuntimeError: TT_FATAL @ ../tt_metal/impl/buffers/buffer.cpp:38: valid_page_size
# For valid non-interleaved buffers page size 2048 must equal buffer size X. For interleaved-buffers page size should be divisible by buffer size
def HasValidPageSize(shape, strict=False):
    if len(shape) >= 2 and shape[-1] > 0:
        return shape[-1] % 32 == 0 or (not strict and shape[-1] < 32)
    return False


# Ttnn globals added with torch.fx._register_custom_builtin
class TtnnDevice:
    def __repr__(self):
        return f"ttnn_Specified_Device"


class TtnnRowMajorLayout:
    def __repr__(self):
        return f"ttnn_ROW_MAJOR_LAYOUT"


class TtnnTileLayout:
    def __repr__(self):
        return f"ttnn_TILE_LAYOUT"


class TtnnInt32:
    def __repr__(self):
        return f"ttnn_int32"


class TtnnUint32:
    def __repr__(self):
        return f"ttnn_uint32"


class TtnnBfloat16:
    def __repr__(self):
        return f"ttnn_bfloat16"


class TtnnDramMemoryConfig:
    def __repr__(self):
        return f"ttnn_DRAM_MEMORY_CONFIG"


class TtnnL1MemoryConfig:
    def __repr__(self):
        return f"ttnn_L1_MEMORY_CONFIG"


# repr_str -> (object_wrapper, object)
__custom_objects_registry = {}


# Note: key MUST be unique!
def get_add_custom_object_in_graph(key: str, obj):
    """
    Register a custom object in the FX graph with a unique string representation.

    This function creates a wrapper object that represents the given object in the FX graph.
    The wrapper's __repr__ returns the provided key, ensuring the object can be properly
    serialized and deserialized within the graph.

    Args:
        key (str): A unique string identifier for the object in the graph.
        obj: The actual object to be wrapped and registered.

    Returns:
        WrapperObj: A wrapper object that represents the original object in the FX graph.

    Side effects:
        - Registers the key as a custom builtin in torch.fx.graph
        - Adds the wrapper and original object to __custom_objects_registry
    """
    if key in __custom_objects_registry:
        return __custom_objects_registry[key][0]

    class WrapperObj:
        def __repr__(self):
            return key

    torch.fx.graph._register_custom_builtin(key, "", obj)
    __custom_objects_registry[key] = (WrapperObj(), obj)
    return __custom_objects_registry[key][0]


def get_emplace_custom_object_in_graph(object_type, *args, **kwargs):
    """
    Create an instance of object_type and register it in the FX graph with a unique name.

    This function creates a unique string identifier based on the object's class name and parameters (args, kwargs),
    then instantiates the object and registers it in the graph using get_add_custom_object_in_graph.

    Args:
        object_type: The class to instantiate
        *args: Positional arguments to pass to the object's constructor
        **kwargs: Keyword arguments to pass to the object's constructor

    Returns:
        WrapperObj: A wrapper object that represents the instantiated object in the FX graph.
        You should use this object when passing to ttnn ops.

    Example:
        >>> program_config = get_emplace_custom_object_in_graph(
        >>>     ttnn.SDPAProgramConfig,
        >>>     compute_with_storage_grid_size=device.compute_with_storage_grid_size(),
        >>>     q_chunk_size=default_q_chunk_size,
        >>>     k_chunk_size=default_k_chunk_size,
        >>>     exp_approx_mode=False,
        >>> )
        >>> res_node = g.call_function(ttnn.transformer.scaled_dot_product_attention, args, kwargs={"program_config": program_config})

    Side effects:
        - Registers the key as a custom builtin in torch.fx.graph
        - Adds the wrapper and original object to __custom_objects_registry
    """

    def sanitize(s):
        # Replace non-alphanumeric characters with underscores
        s = re.sub(r"[^0-9a-zA-Z_]", "_", str(s))
        # Remove consecutive underscores
        s = re.sub(r"_+", "_", s)
        return s.strip("_")

    # This representation of an object MUST be unique
    def build_repr():
        repr_str = f"{object_type.__name__}"
        for arg in args:
            repr_str += f"_{sanitize(arg)}"
        for k, v in kwargs.items():
            repr_str += f"_{sanitize(k)}_{sanitize(v)}"
        return repr_str

    key = build_repr()

    return get_add_custom_object_in_graph(key, object_type(*args, **kwargs))



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import os
from setuptools import setup, find_packages


def read_readme():
    with open("README.md", encoding="utf-8") as f:
        return f.read()


setup(
    name="torch-ttnn",
    description="PyTorch 2.0 TTNN Compiler",
    long_description=read_readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/tenstorrent/pytorch2.0_ttnn",
    author="Tenstorrent",
    author_email="info@tenstorrent.com",
    license="Apache-2.0",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Operating System :: POSIX :: Linux",
    ],
    packages=find_packages(exclude=["tests", "tests.*"]),
    python_requires=">=3.10",
    install_requires=[
        "torch==2.2.1+cpu",
        "torchvision==0.17.1+cpu",
        "tabulate==0.9.0",
        "networkx==3.1",
        "graphviz",
        "matplotlib==3.7.1",
        "paramiko==3.5.1",
    ],
    extras_require={
        "ttnn": ["ttnn==0.59.0"],
        "dev": [
            "pytest>=7.4.0",
            "pytest-cov>=4.1.0",
            "black>=23.7.0",
            "isort>=5.12.0",
            "flake8>=6.1.0",
            "mypy>=1.5.0",
            "pre-commit>=3.3.3",
        ],
    },
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
)



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
from torch_ttnn.passes.lowering.add_data_move_pass import is_tt_data_move, is_tt_compute
from torch_ttnn.utils import TtnnDevice
import ttnn
import torch

# Configs
device_name = "wormhole"
cores = 120
L1_mem = 1048576  # 1 MB
circular_buffer = 20 * 1048576  # 20 MB
# SRAM_LIMIT = cores * L1_mem - circular_buffer (100 MB in this case)


# This will manage all memory related operations & data
class MemoryManager:
    def __init__(self, device, cores, L1_mem, circular_buffer, verbose=True):
        # Config
        self.device = device
        self.cores = cores
        self.L1_mem = L1_mem
        self.circular_buffer = circular_buffer
        self.verbose = verbose
        self.total_sram_limit = self.cores * self.L1_mem
        self.free_sram = self.total_sram_limit - self.circular_buffer
        self.usable_sram_limit = self.total_sram_limit - self.circular_buffer
        self.used_sram = 0

        # Tensor memory address allocation tracker
        self.last_assigned_addr = 0
        # Tracks tensors currently in SRAM memory
        self.tensors_on_device = []
        # Each ttnn op's input & output tensor ids
        self.ttnn_ops_tids = {}
        # Current usage of SRAM memory at each step of ttnn op execution timeline
        self.sram_usage_at = {}
        # Keep tally of tensors in SRAM memory at each step of ttnn op execution timeline
        self.tids_in_sram_at = {}
        # Tensor size of respective tensors
        self.tensor_size_of = {}
        # Map of each node with its output tensor id
        self.node_to_tid_map = {}
        # Sum of input & output tensor sizes for each ttnn op
        self.ops_mem_usage = {}
        # Memory address of each tensor id in SRAM
        self.tid_to_addr_map_in_sram = {}
        # Maximum usage of SRAM during lifetime of model execution
        self.peak_sram_usage = 0
        # Data for plotting on a chart
        self.data_points = {}
        # Logs generated during memory analysis
        self.logs = ""

    def has_sram_overflow(self):
        for _, sram_usage in self.sram_usage_at.items():
            if sram_usage >= self.usable_sram_limit:
                return True
        return False

    def set_peak_sram_usage(self):
        peak_sram_usage = 0
        for _, sram_usage in self.sram_usage_at.items():
            if sram_usage > peak_sram_usage:
                peak_sram_usage = sram_usage
        self.peak_sram_usage = peak_sram_usage

    def max_evictions_required(self):
        max_evictions = 0
        for _, sram_usage in self.sram_usage_at.items():
            if sram_usage >= self.usable_sram_limit:
                max_evictions += 1
        return max_evictions

    def log(self, message: str):
        self.logs += message
        if self.verbose:
            print(message)


# This holds op related information one can query from
class OpRegistry:
    def get_tensor_shape_and_dtype(self, node):
        if is_tt_data_move(node):
            assert len(node.all_input_nodes) == 1, "Data movement operators can't have more than one input!"
            return self.get_tensor_shape_and_dtype(node.all_input_nodes[0])
        else:
            # TODO: What if meta of nth output of the node is requested?
            if isinstance(node.meta["val"], tuple):
                return (node.meta["val"][0].size(), node.meta["val"][0].dtype)
            else:
                return (node.meta["val"].size(), node.meta["val"].dtype)

    def is_input_tensor_on_device(self, node):
        if node.target == ttnn.from_torch:
            if "device" in node.kwargs and isinstance(node.kwargs["device"], TtnnDevice):
                return True
        if node.target is ttnn.to_device or is_tt_compute(node):
            return True
        # This is a data move op which propagates shape of input tensor
        if node.target == ttnn.Shape or node.target == ttnn.from_device:
            return False
        # Considers cases like ttnn op --> to_layout
        if is_tt_data_move(node):
            return self.is_input_tensor_on_device(node.all_input_nodes[0])
        else:
            return False


def check_sram_overflow(mm: MemoryManager) -> bool:
    return mm.has_sram_overflow()


def which_tensors_to_evict(mm: MemoryManager) -> tuple:
    sram_usage = 0
    tensors_to_evict = []
    potential_tensors_to_evict = []
    for node_name, sram_used in mm.sram_usage_at.items():
        if sram_used >= mm.usable_sram_limit:
            sram_usage = sram_used

            # Checks if the ttnn op itself can fit in SRAM or not.
            if mm.ops_mem_usage[node_name] >= mm.usable_sram_limit:
                print(f"Splitting required, {node_name} op doesn't fit in SRAM!")
                return (node_name, -1)

            # Removes tids of current ttnn op from list of potential tensors that can be evicted
            current_op_tids = mm.ttnn_ops_tids[node_name]
            potential_tensors_to_evict = mm.tids_in_sram_at[node_name]
            for tid in current_op_tids:
                if tid in potential_tensors_to_evict:
                    potential_tensors_to_evict.remove(tid)

            # Convert list of potential tensors to evict to dict with tensor's memory size
            potential_tensors_to_evict = {tid: mm.tensor_size_of[tid] for tid in potential_tensors_to_evict}
            # Sort the dict as per tensor memory size (in decreasing order)
            potential_tensors_to_evict = {
                k: v
                for k, v in sorted(
                    potential_tensors_to_evict.items(),
                    key=lambda item: item[1],
                    reverse=True,
                )
            }
            break

    if len(potential_tensors_to_evict) > 0:
        for tid, tensor_size in potential_tensors_to_evict.items():
            sram_usage -= tensor_size
            tensors_to_evict.append(tid)
            if sram_usage < mm.usable_sram_limit:
                break

    return (node_name, tensors_to_evict)


def get_dtype_size(dtype):
    if dtype in [torch.float, torch.float32, torch.int, torch.int32]:
        return 4
    elif dtype in [torch.float64, torch.double, torch.int64, torch.long]:
        return 8
    elif dtype in [torch.float16, torch.half, torch.bfloat16, torch.int16, torch.short]:
        return 2
    else:
        assert False, "Invalid datatype! This is not supported yet."


def get_tensor_size(shape, dtype):
    size = get_dtype_size(dtype)
    for val in list(shape):
        size = val * size
    return size



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import torch
from torch._subclasses.fake_tensor import unset_fake_temporarily
from torch.fx.passes.infra.pass_base import PassBase, PassResult
from .lowering.to_tt_guard import can_lowering_to_ttnn


class ConstantFoldingPass(PassBase):
    def __init__(self):
        super().__init__()
        self.foldable_ops = {
            torch.ops.aten.lift_fresh_copy.default,
            torch.ops.aten.pow.Tensor_Tensor,
            torch.ops.aten.arange.start,
            torch.ops.aten.arange.start_step,
            torch.ops.aten.unsqueeze.default,
            torch.ops.aten.arange.default,
            torch.ops.aten.view.default,
            torch.ops.aten.add.Tensor,
            torch.ops.aten.mul.Tensor,
            torch.ops.aten._to_copy.default,
            torch.ops.aten.expand.default,
            torch.ops.aten.sub.Tensor,
            torch.ops.aten.ceil.default,
            torch.ops.aten.clamp.default,
            torch.ops.aten.ones.default,
            torch.ops.aten.cumsum.default,
            torch.ops.aten._unsafe_index.Tensor,
            torch.ops.aten.index.Tensor,
            torch.ops.aten.index_select.default,
            torch.ops.aten.ne.Scalar,
            torch.ops.aten.select.int,
            torch.ops.aten.bitwise_not.default,
            torch.ops.aten.floor_divide.default,
        }

    def call(self, gm: torch.fx.GraphModule):
        with unset_fake_temporarily():
            for node in gm.graph.nodes:
                if node.op == "call_function" and node.target in self.foldable_ops and self._can_fold(gm, node):
                    try:
                        folded_value = self._evaluate_node(gm, node)
                        self._replace_with_constant(gm, node, folded_value)
                    except Exception as e:
                        print(f"Warning: Could not fold node {node.name}: {e}")

        return PassResult(gm, True)

    def _can_fold(self, gm: torch.fx.GraphModule, node):
        if not can_lowering_to_ttnn(node):
            return False

        def can_arg_fold(arg):
            if (
                isinstance(
                    arg,
                    (
                        int,
                        float,
                    ),
                )
                or arg is None
            ):
                return True

            if isinstance(arg, torch.fx.Node) and arg.op in ("get_attr", "constant"):
                return True
            return False

        for arg in node.args:
            if isinstance(
                arg,
                (
                    list,
                    tuple,
                    torch.fx.immutable_collections.immutable_list,
                ),
            ):
                for arg_element in arg:
                    if not can_arg_fold(arg_element):
                        return False
            elif not can_arg_fold(arg):
                return False

        return True

    def _evaluate_node(self, gm: torch.fx.GraphModule, node):
        def eval_arg(arg):
            if isinstance(arg, torch.fx.Node):
                if arg.op == "get_attr":
                    return getattr(gm, arg.target)
                elif arg.op == "constant":
                    return arg.args[0]
            return arg

        args = []
        for arg in node.args:
            if isinstance(
                arg,
                (
                    list,
                    tuple,
                    torch.fx.immutable_collections.immutable_list,
                ),
            ):
                list_arg = []
                for arg_element in arg:
                    list_arg.append(eval_arg(arg_element))
                args.append(list_arg)
            else:
                args.append(eval_arg(arg))

        if node.target == torch.ops.aten.lift_fresh_copy.default:
            return args[0]

        if node.target in self.foldable_ops:
            result = node.target(*args, **node.kwargs)
            return result

        # Add handlers for other operations...

        raise ValueError(f"Unsupported operation: {node.target}")

    def _replace_with_constant(self, gm: torch.fx.GraphModule, node, value):
        with gm.graph.inserting_before(node):
            new_node = gm.graph.create_node("get_attr", target=f"_folded_{node.name}", args=(), kwargs=None)
            # Copying all of the meta causes some ops to be missing
            new_node.meta["seq_nr"] = node.meta["seq_nr"]
            new_node.meta["original_aten"] = node.meta["original_aten"]
            new_node.meta["tensor_meta"] = node.meta["tensor_meta"]

        gm.register_parameter(f"_folded_{node.name}", torch.nn.Parameter(value, requires_grad=False))

        node.replace_all_uses_with(new_node)
        gm.graph.erase_node(node)



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
from functools import partial
from .to_tt_guard_autogen import *

############################################################
# EXTRA BLOCKLIST OF Whisper
############################################################
# torch._dynamo.exc.BackendCompilerFailed: backend='ttnn_backend' raised:
# RuntimeError: aten::clone() Expected a value of type 'Tensor' for argument 'self' but instead found type 'SymInt'.
# Position: 0
# Value: s1
# Declaration: aten::clone(Tensor self, *, MemoryFormat? memory_format=None) -> Tensor
# Cast error details: Unable to cast s1 to Tensor
# TODO: not pass yet

############################################################
# EXTRA BLOCKLIST OF GPTNeo
############################################################
# torch._dynamo.exc.BackendCompilerFailed: backend='ttnn_backend' raised:
# RuntimeError: aten::clone() Expected a value of type 'Tensor' for argument 'self' but instead found type 'SymInt'.
# Position: 0
# Value: s0
# Declaration: aten::clone(Tensor self, *, MemoryFormat? memory_format=None) -> Tensor
# Cast error details: Unable to cast s0 to Tensor
# TODO: not pass yet

############################################################
# EXTRA BLOCKLIST OF codegen
############################################################
# ttnn.from_torch(torch.tensor(False)) => ttnn.gt
# ttnn.from_torch not support sclar of bool dtype
# RuntimeError: TT_FATAL @ tensor/types.cpp:209: normalized_index >= 0 and normalized_index < rank
# not lowering ttnn.gt to avoid ttnn.from_torch of sclar of bool dtype
aten_gt_Scalar_blocklist = [["Tensor<[]> self = ?", "number other = 0"]]

# torch._dynamo.exc.BackendCompilerFailed: backend='ttnn_backend' raised:
# RuntimeError: aten::clone() Expected a value of type 'Tensor' for argument 'self' but instead found type 'SymInt'.
# Position: 0
# Value: s0
# Declaration: aten::clone(Tensor self, *, MemoryFormat? memory_format=None) -> Tensor
# Cast error details: Unable to cast s0 to Tensor
# TODO: not pass yet

############################################################
# EXTRA BLOCKLIST OF OPT
############################################################
# ttnn.from_torch(torch.tensor(-3.3895e+38)) => ttnn.maximum
# ttnn.from_torch not support scalar
# RuntimeError: TT_FATAL @ tensor/types.cpp:209: normalized_index >= 0 and normalized_index < rank
# not lowering ttnn.maximum to avoid ttnn.from_torch of scalar
# aten_maximum_default_blocklist += [["Tensor<[1, 16, 59, 59]> self = ?", "Tensor other = ?"]]

# torch._dynamo.exc.BackendCompilerFailed: backend='ttnn_backend' raised:
# RuntimeError: aten::clone() Expected a value of type 'Tensor' for argument 'self' but instead found type 'SymInt'.
# Position: 0
# Value: s0
# Declaration: aten::clone(Tensor self, *, MemoryFormat? memory_format=None) -> Tensor
# Cast error details: Unable to cast s0 to Tensor
# TODO: not pass yet

############################################################
# EXTRA BLOCKLIST OF t5
############################################################
# torch._dynamo.exc.BackendCompilerFailed: backend='ttnn_backend' raised:
# Exception: unhashable type: non-singleton SymInt
# TODO: not pass yet

############################################################
# EXTRA BLOCKLIST OF DETR
############################################################
# TypeError: __call__(): incompatible function arguments. The following argument types are supported:
#     1. (self: ttnn._ttnn.operations.moreh.moreh_cumsum_t, input: ttnn._ttnn.tensor.Tensor, dim: int, *, output: Optional[ttnn._ttnn.tensor.Tensor] = None, memory_config: Optional[ttnn._ttnn.tensor.MemoryConfig] = None, compute_kernel_config: Optional[Union[ttnn._ttnn.operations.core.GrayskullComputeKernelConfig, ttnn._ttnn.operations.core.WormholeComputeKernelConfig]] = None) -> ttnn._ttnn.tensor.Tensor
#
# Invoked with: <ttnn._ttnn.operations.moreh.moreh_cumsum_t object at 0x7faef0b78b30>, ttnn.Tensor([[[[ 0.00000,  0.00000,  ...,  0.00000,  0.00000],
#               ...,
#                [ 0.00000,  0.00000,  ...,  0.00000,  0.00000]]]], shape=Shape([1, 23, 40[64], 1[32]]), dtype=DataType::BFLOAT16, layout=Layout::TILE), 1; kwargs: dtype=torch.float32
aten_cumsum_default_blocklist = [["Tensor<[1, 23, 40]> self = ?", "int dim = 1", "Optional[int] dtype = torch.float32"]]

############################################################
# EXTRA BLOCKLIST OF retinanet_resnet50_fpn
############################################################
# Statically allocated circular buffers on core range [(x=0,y=0) - (x=0,y=0)] grow to 3580704 B which is beyond max L1 size of 1499136 B
aten_aten_stack_default = [
    ["List[Tensor] tensors = [<[13600]>, <[13600]>, <[13600]>, <[13600]>]", "int dim = 1"],
]

############################################################
# EXTRA BLOCKLIST OF falcon-7b-instruct
############################################################
# RuntimeError: TT_THROW @ /tmp/build-via-sdist-d26xvola/ttnn-0.54.0rc18+wormhole.b0/ttnn/cpp/ttnn/device_operation.hpp:487: tt::exception
# Unsupported storage type
aten_argmax_default_blocklist += [["Tensor<[1, 7]> self = ?", "Optional[int] dim = 1", "bool keepdim = True"]]

############################################################
# EXTRA BLOCKLIST OF retinanet_resnet50_fpn_v2
############################################################
# Statically allocated circular buffers on core range [(x=0,y=0) - (x=0,y=0)] grow to 3580704 B which is beyond max L1 size of 1499136 B
# TODO: not pass yet

############################################################

GUARD[torch.ops.aten.gt.Scalar] = partial(guard_aten, aten_gt_Scalar_blocklist)
GUARD[torch.ops.aten.cumsum.default] = partial(guard_aten, aten_cumsum_default_blocklist)
GUARD[torch.ops.aten.stack.default] = partial(guard_aten, aten_aten_stack_default)


def can_lowering_to_ttnn(node):
    if node.target in GUARD:
        return GUARD[node.target](node)

    return True



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0

import torch
from torch.fx.passes.infra.pass_base import PassBase, PassResult
from torch_ttnn.passes.patterns.linear_pattern import LinearPatterns
from torch_ttnn.passes.patterns.soft_max_pattern import SoftMaxPatterns
from torch_ttnn.passes.patterns.attention_pattern import AttentionPatterns
from pathlib import Path


class FusionPass(PassBase):
    def __init__(self):
        super().__init__()

    def call(self, gm: torch.fx.GraphModule):
        modified = False

        # List of pattern classes to apply
        pattern_classes = [LinearPatterns, SoftMaxPatterns]
        # AttentionPatterns must be applied last
        pattern_classes.append(AttentionPatterns)

        for pattern_cls in pattern_classes:
            pattern = pattern_cls(gm)
            matches = pattern.match_pattern()
            pattern.replace_pattern(matches)
            modified |= len(matches) > 0

        return PassResult(gm, modified)
